import asyncio
import datetime
import json
import math
import time
from binascii import unhexlify

import serial
from loguru import logger

from ..interfaces import database
from ..utils import config, bluetooth, multiprocessing


class Parameter:
    def __init__(self, data, field, start_byte, stop_byte, byte_order='big', signed=False, precision=1.0):
        self.field = field
        self.start_byte = start_byte
        self.stop_byte = stop_byte
        self.byte_order = byte_order
        self.signed = signed
        self.precision = precision
        self.data = data[start_byte: stop_byte]
        self.value = int.from_bytes(data[self.start_byte: self.stop_byte], byteorder=self.byte_order,
                                    signed=self.signed) * self.precision
        if self.precision < 1:
            self.value = round(self.value, math.ceil(math.log(1 / self.precision, 10)))

    def get_bits(self, index=None):
        string_bytes = self.data
        bytes_list = []
        for i in range(len(string_bytes)):
            bytes_list.append(string_bytes[i:i + 1])

        bits = []
        for byte in bytes_list:
            for i in range(8):
                bits.append((ord(byte) >> i) & 1)
        return bits


def initialize(main_task_queue):
    try:
        ser = serial.Serial(
            port=config.homesolar_config['BLUETOOTH']['port'],
            baudrate=9600,
            parity=serial.PARITY_NONE,
            stopbits=serial.STOPBITS_ONE,
            bytesize=serial.EIGHTBITS)

        try:
            ser.open()
            time.sleep(1)
        except Exception as e:
            ser.close()
            ser.open()
            logger.warning(f"Something went wrong when trying to open Bluetooth as a Serial Connection [{e}]")

        first_boot = True
        while True:
            try:
                ser.write(unhexlify(config.homesolar_config['BLUETOOTH']['request_code']))
                if first_boot:
                    time.sleep(1)
                    first_boot = False
                antw33 = ser.read(140)

                params = [Parameter(antw33, "Voltage", 4, 6, 'big', True, 0.1),
                          Parameter(antw33, "Current", 70, 74, 'big', True, 0.1),
                          Parameter(antw33, "Power", 111, 115, 'big', True),
                          Parameter(antw33, "Capacity", 75, 79, 'big', False, 0.000001),
                          Parameter(antw33, "ChargeStatus", 103, 104), Parameter(antw33, "DischargeStatus", 104, 105),
                          Parameter(antw33, "BalanceStatus", 105, 106), Parameter(antw33, "BalanceCells", 133, 136),
                          Parameter(antw33, "Temp1", 91, 93), Parameter(antw33, "Temp2", 93, 95),
                          Parameter(antw33, "Temp3", 95, 97), Parameter(antw33, "Temp4", 97, 99),
                          Parameter(antw33, "Temp5", 99, 101), Parameter(antw33, "Temp6", 101, 103),
                          Parameter(antw33, "MinCell", 118, 119), Parameter(antw33, "MaxCell", 115, 116),
                          Parameter(antw33, "MinVolt", 119, 121, 'big', False, 0.001),
                          Parameter(antw33, "MaxVolt", 116, 118, 'big', False, 0.001),
                          Parameter(antw33, "AvgVolt", 121, 123, precision=0.001),
                          Parameter(antw33, "Cell1", 6, 8, 'big', False, 0.001),
                          Parameter(antw33, "Cell2", 8, 10, 'big', False, 0.001),
                          Parameter(antw33, "Cell3", 10, 12, 'big', False, 0.001),
                          Parameter(antw33, "Cell4", 12, 14, 'big', False, 0.001),
                          Parameter(antw33, "Cell5", 14, 16, 'big', False, 0.001),
                          Parameter(antw33, "Cell6", 16, 18, 'big', False, 0.001),
                          Parameter(antw33, "Cell7", 18, 20, 'big', False, 0.001),
                          Parameter(antw33, "Cell8", 20, 22, 'big', False, 0.001),
                          Parameter(antw33, "Cell9", 22, 24, 'big', False, 0.001),
                          Parameter(antw33, "Cell10", 24, 26, 'big', False, 0.001),
                          Parameter(antw33, "Cell11", 26, 28, 'big', False, 0.001),
                          Parameter(antw33, "Cell12", 28, 30, 'big', False, 0.001),
                          Parameter(antw33, "Cell13", 30, 32, 'big', False, 0.001),
                          Parameter(antw33, "Cell14", 32, 34, 'big', False, 0.001),
                          Parameter(antw33, "Cell15", 34, 36, 'big', False, 0.001),
                          Parameter(antw33, "Cell16", 36, 38, 'big', False, 0.001),
                          Parameter(antw33, "Cell17", 38, 40, 'big', False, 0.001),
                          Parameter(antw33, "Cell18", 40, 42, 'big', False, 0.001),
                          Parameter(antw33, "Cell19", 42, 44, 'big', False, 0.001),
                          Parameter(antw33, "Cell20", 44, 46, 'big', False, 0.001),
                          Parameter(antw33, "Cell21", 46, 48, 'big', False, 0.001),
                          Parameter(antw33, "Cell22", 48, 50, 'big', False, 0.001),
                          Parameter(antw33, "Cell23", 50, 52, 'big', False, 0.001),
                          Parameter(antw33, "Cell24", 52, 54, 'big', False, 0.001)]

                voltage = 0
                fields = {}
                for data in params:
                    fields[data.field] = data.value

                    if data.field == 'AvgVolt':
                        voltage = data.value

                    if data.field == 'BalanceCells':
                        for index, bal in enumerate(data.get_bits()):
                            fields[f"Bal{index}"] = bal

                    if data.field == 'AvgVolt':
                        fields['SoC'] = bluetooth.ManualSoC().get_soc(data.value)

                if voltage == 0:
                    ser.close()
                    time.sleep(1)
                    ser.open()
                    time.sleep(1)

                data = {
                    "name": "Antw33-BMS",
                    "payload": json.dumps(fields),
                    'time': datetime.datetime.now().timestamp()
                }
                logger.debug(f"Incoming Sensor Data [{data['name']}]")
                task = {
                    "name": "write_sensor_data",
                    "data": data
                }
                main_task_queue.put(task)
                time.sleep(5)
            except:
                ser.close()
                time.sleep(1)
                ser.open()
                time.sleep(1)

    except Exception as ex:
        print(ex)
    finally:
        try:
            ser.close()
        except Exception as e:
            logger.warning(f"Something went wrong when trying to close Bluetooth as Serial Connection [{e}]")



import asyncio
from loguru import logger

from ..interfaces import database
from ..interfaces import mqtt
from ..services import mqtt as mqtt_services

POISON_PILL = "STOP"
homesolar_database_queue = None
homesolar_task_queue = None


def load_queue(queue1, queue2):
    global homesolar_database_queue, homesolar_task_queue
    homesolar_database_queue = queue1
    homesolar_task_queue = queue2


# def process_task(queue):
#     try:
#         while True:
#             new_data = queue.get()
#
#             if new_data == POISON_PILL:
#                 break
#
#             asyncio.run(database.write_sensor_data(new_data))
#     except:
#         pass
#     finally:
#         return


def process_task(queue):
    try:
        while True:
            logger.debug("Waiting for task")
            task = queue.get()

            logger.debug(f"Task received [{task}]")

            if task == POISON_PILL:
                break

            try:
                if task["name"] == "write_sensor_data":
                    asyncio.run(database.write_sensor_data(task["data"]))
                elif task["name"] == "send_summary":
                    mqtt.send_summary(mqtt_services.mqtt_client, task["update"])
                elif task["name"] == "send_battery":
                    mqtt.send_battery(mqtt_services.mqtt_client)
                elif task["name"] == "chart_data":
                    database.get_chart_data(task["date"], task["timescale"])
                elif task["name"] == "update_config":
                    pass
                else:
                    logger.warning(f"Invalid task, discarding.. [{task}]")
            except Exception as e:
                logger.exception(f"Something went wrong when trying to do a task [{e}]")
    except:
        pass
    finally:
        return



# Third party imports
import datetime
import json
import threading

from loguru import logger
from paho.mqtt import client as paho_mqtt

# Project imports
from ..utils.mqtt import MqttTopic


# This thread is responsible for any task that needed to be executed on mqtt process
# mainly for publish message task and several other task for internal commands.
# The reason it has to be on a different thread is to make sure that whatever task that is processed
# would not hog the MQTT client ability to received message from the broker
# and skipping any information that might be important
def mqtt_task_loop(run, queue, client: paho_mqtt.Client):
    try:
        while True:
            # Waiting for task to do
            logger.debug("Waiting for task!")
            task = queue.get()
            logger.debug(f"Task received [{task}]")

            # If stop is issued, shutting down the thread
            if task == "STOP":
                logger.info("Stop is issued, shutting down the thread")
                break

            try:
                if task["name"] == "publish":
                    client.publish(task["topic"], task["payload"])
                else:
                    logger.warning(f"Unknown task is issued, discarding... [{task}]")
            except Exception as e:
                logger.warning(f"Invalid task is issued, discarding... [{e}]")

    except Exception as e:
        logger.exception(f"Unexpected error occurred, shutting down thread [{e}]")
    finally:
        run.append("STOP")
        return


# Start the MQTT Service, designed to only work properly on an entirely different process
# Any other implementation and usage of the service is not recommended, only do so if you understand what you're doing
def start_service(mqtt_queue, main_queue, config):
    # Initialize the client and thread
    run = []
    client = paho_mqtt.Client()
    mqtt_task_thread = threading.Thread(target=mqtt_task_loop, args=[run, mqtt_queue, client])

    try:
        host = config['MQTT']['host']
        port = config['MQTT']['port']
        username = config['MQTT']['username']
        password = config['MQTT']['password']
        keepalive = config['MQTT']['keepalive']
        reconnect = config['MQTT']['reconnect_delay']

        client.reconnect_delay_set(max_delay=reconnect)
        client.on_message = decorated_on_message(mqtt_queue, main_queue)
        client.on_connect = on_connect
        client.username_pw_set(username, password)
        client.connect(host, port, keepalive)

        # Start the task thread
        mqtt_task_thread.start()
        logger.info("Publish thread started")

        # Start MQTT main loop
        client.loop_forever()

    except Exception as e:
        logger.exception(f"Something unexpected happened when running MQTT Service, shutting down... [{e}]")
    finally:
        client.disconnect()
        mqtt_queue.put("STOP")
        mqtt_task_thread.join()
        main_queue.put("STOP")


def decorated_on_message(mqtt_queue, main_queue):
    def on_message(client: paho_mqtt.Client, userdata, msg):
        # Handles messages on Sensors Topic
        if MqttTopic.SENSORS[:-2] in str(msg.topic):
            sensor_name = str(msg.topic).replace(MqttTopic.SENSORS[:-2] + "/", "").rstrip('/')
            data = {
                "name": sensor_name,
                "payload": msg.payload,
                "time": datetime.datetime.now().timestamp()
            }

            logger.debug(f"Incoming Sensor Data [{sensor_name}]")
            main_queue.put({"name": "write_sensor_data", "data": data})

        # Handles messages on Client Topic
        elif MqttTopic.CLIENT[:-2] in str(msg.topic):
            client_id = str(msg.topic).split("/")[-2]
            logger.debug(f"Client [{client_id}] request incoming {msg.payload}")

            # Handles Client status update
            if "status" in str(msg.topic):
                logger.debug("Client update request received")
                status = str(msg.payload, 'utf-8')
                try:
                    if status == "Connected":
                        task = {
                            "name": "add_client",
                            "client_id": client_id
                        }
                    else:
                        task = {
                            "name": "remove_client",
                            "client_id": client_id
                        }
                    main_queue.put(task)

                except Exception as e:
                    logger.warning(f"Something went wrong when updating client's status {e}")

            # Handles client request
            elif "request" in str(msg.topic):
                logger.debug("Client request received")
                try:
                    json_data = json.loads(msg.payload)
                    request_type = json_data["request_type"]
                    task = {
                        "name": request_type,
                        "request_id": json_data["request_id"],
                        "client_id": client_id
                    }

                    if request_type == "chart_data":
                        task["date"] = json_data["date"]
                        task["timescale"] = json_data["timescale"]
                        main_queue.put(task)
                    else:
                        logger.warning(f"Unknown request is issued, discarding... [{request_type}]")

                except Exception as e:
                    logger.warning(f"Invalid request, rejected [{e}]")
            else:
                logger.warning(f"Unknown request received, discarding... [{msg.topic}:{msg.payload}]")

    return on_message


def on_connect(client: paho_mqtt.Client, userdata, flags, reconnect):
    try:
        client.subscribe(MqttTopic.SENSORS)
        client.subscribe(MqttTopic.REQUEST)
        client.subscribe(MqttTopic.CLIENT)
        message = "Client reconnected successfully" if reconnect else "Client connected successfully"
        logger.info(message)
    except Exception as e:
        logger.exception(f"Something went wrong while trying to subscribes {e}")



import numpy as np
from scipy import interpolate

# Status MOSFET saat Discharge
MOSFET_Discharge_St = ["OFF", "ON", "Cell Overdischarge", "Overcurrent", "4-Unknown", "Pack Overdischarge",
                       "Battery Overheat", "MOSFET Overheat", "Abnormal Current", "Battery Not Detected",
                       "PCB Overheat", "Charge MOSFET Turned On", "Shortcircuit", "Discharge MOSFET Abnormality",
                       "Start exception", "Manual OFF"]
# Status MOSFET saat Charge
MOSFET_Charge_St = ["OFF", "ON", "Cell Overcharge", "Overcurrent", "Battery Full", "Pack Overvoltage",
                    "Battery Overheat", "MOSFET Overheat", "Abnormal Current", "Battery Not Detected",
                    "PCB Overheat", "11-Unknown", "12-Unknown", "Discharge MOSFET Abnormality", "14", "Manual OFF"]
# Status Balancing
Bal_St = ["OFF", "Exceeding Trigger Limit", "Voltage Diff Too High", "Overheat", "ACTIVE",
          "5-Unknown", "6-Unknown", "7-Unknown", "8-Unknown", "9-Unknown", "PCB Overheat"]


def find_nearest(array, value):
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return idx


class ManualSoC:
    y = [2.8, 3.6, 3.7, 3.72, 3.75, 3.78, 3.8, 3.9, 3.95, 4.0, 4.2]
    x = [0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100]

    xarr = np.array(x)
    yarr = np.array(y)

    splines = interpolate.splrep(xarr, yarr)
    x_vals = np.linspace(xarr.min(), xarr.max(), 101 * 10)
    y_vals = interpolate.splev(x_vals, splines)

    def get_soc(self, volt):
        x = find_nearest(self.y_vals, volt)
        return self.x_vals[x]



import asyncio
import json
from time import perf_counter

from loguru import logger

# project's imports
from ..services import influxdb, sqlite
from ..utils import datetime, config, data_formatter
from ..utils import influxdb as influx_utils
from ..utils import sqlite as sqlite_utils
from ..utils.sqlite import SensorData


# Used on All database
async def write_sensor_data(data):
    start_time = perf_counter()
    sensor_data = data_formatter.format_sensor_data(data)
    await influxdb.write(sensor_data)
    logger.debug("is it waiting ?")
    mapped_data = sqlite_utils.mapped_for_upsert(sensor_data)
    await sqlite.bulk_upsert_sensors(mapped_data)
    logger.debug(json.dumps(sensor_data))
    logger.debug(f"Time taken for entire sensor_data_write: {perf_counter() - start_time} second(s)")


async def write_sensor_to_influxdb(data):
    await influxdb.write(data_formatter.format_sensor_data(data))


async def write_sensor_to_sqlite(data):
    await sqlite.bulk_upsert_sensors(sqlite_utils.mapped_for_upsert(data_formatter.format_sensor_data(data)))


# Used only on InfluxDB
async def get_battery_charge(date):
    flux = influx_utils.generate_flux(
        config.homesolar_config['DATA']['battery_charge_measurement'],
        config.homesolar_config['DATA']['battery_charge_field'],
        datetime.stringify_timestamp(date),
        datetime.get_next_day(date), "DAY"
    )
    result = await influxdb.query(flux)
    logger.debug(result)
    return data_formatter.simplify_serialized_data(influx_utils.serialize(result, "DAY"))


async def get_battery_usage(date, timescale):
    start_time, stop_time = datetime.get_date_pair(date, timescale)
    flux = influx_utils.generate_flux(
        config.homesolar_config['DATA']['battery_power_measurement'],
        config.homesolar_config['DATA']['battery_power_field'],
        start_time,
        stop_time,
        timescale
    )
    result = await influxdb.query(flux)
    logger.debug(result)
    return data_formatter.simplify_serialized_data(influx_utils.serialize(result, timescale))


async def get_solar_production(date, timescale):
    start_time, stop_time = datetime.get_date_pair(date, timescale)
    flux = influx_utils.generate_flux(
        config.homesolar_config['DATA']['solar_production_measurement'],
        config.homesolar_config['DATA']['solar_production_field'],
        start_time,
        stop_time,
        timescale
    )
    result = await influxdb.query(flux)
    logger.debug(result)
    return data_formatter.simplify_serialized_data(influx_utils.serialize(result, timescale))


async def get_grid_usage(date, timescale):
    start_time, stop_time = datetime.get_date_pair(date, timescale)
    flux = influx_utils.generate_flux(
        config.homesolar_config['DATA']['grid_power_measurement'],
        config.homesolar_config['DATA']['grid_power_field'],
        start_time,
        stop_time,
        timescale
    )
    result = await influxdb.query(flux)
    logger.debug(result)
    return data_formatter.simplify_serialized_data(influx_utils.serialize(result, timescale))


async def get_inverter_usage(date, timescale):
    start_time, stop_time = datetime.get_date_pair(date, timescale)
    flux = influx_utils.generate_flux(
        config.homesolar_config['DATA']['inverter_power_measurement'],
        config.homesolar_config['DATA']['inverter_power_field'],
        start_time,
        stop_time,
        timescale
    )
    result = await influxdb.query(flux)
    logger.debug(result)
    return data_formatter.simplify_serialized_data(influx_utils.serialize(result, timescale))


async def get_home_usage(date, timescale):
    start_time, stop_time = datetime.get_date_pair(date, timescale)
    flux = influx_utils.generate_combined_tables_flux(
        [config.homesolar_config['DATA']['grid_power_measurement'],
         config.homesolar_config['DATA']['inverter_power_measurement']],
        [config.homesolar_config['DATA']['grid_power_field'],
         config.homesolar_config['DATA']['inverter_power_field']],
        start_time,
        stop_time,
        timescale
    )
    logger.debug(flux)
    result = await influxdb.query(flux)
    logger.debug(result)
    return data_formatter.simplify_serialized_data(influx_utils.serialize(result, timescale))


async def get_chart_data(date, timescale):
    solar_production = await get_solar_production(date, timescale)
    battery_usage = await get_battery_usage(date, timescale)
    grid_usage = await get_grid_usage(date, timescale)
    inverter_usage = await get_inverter_usage(date, timescale)
    home_usage = await get_home_usage(date, timescale)

    if timescale == "DAY":
        battery_charge = await get_battery_charge(date)
    else:
        battery_charge = None

    data = {
        "solar_production": solar_production,
        "battery_usage": battery_usage,
        "grid_usage": grid_usage,
        "inverter_usage": inverter_usage,
        "home_usage": home_usage,
        "batterY_charge": battery_charge
    }
    return data


# Used only on Sqlite
async def get_sensor_data(measurement, field):
    sensor_data = await sqlite.query(SensorData, SensorData.name.in_([f"{measurement}#{field}"]))
    if sensor_data is None:
        logger.warning("No sensor data with specified name found")
        return None
    else:
        for sensor in sensor_data:
            return data_formatter.parse_to_float_if_possible(sensor.value)



import json
from collections.abc import MutableMapping

from loguru import logger


def parse_to_float_if_possible(string):
    result = string
    try:
        result = float(string)
    except Exception as e:
        logger.warning(f"Cannot parse to float [{e}]")
        result = parse_to_bool_if_possible(string)
    finally:
        return result


def parse_to_bool_if_possible(string):
    if string in ["true", "True"]:
        result = True
    else:
        result = string
        logger.warning("Cannot parse to boolean")
    return result


def flatten_dict(d, parent_key='', sep='_'):
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, MutableMapping):
            items.extend(flatten_dict(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


def format_sensor_data(data):
    try:
        payload = json.loads(data["payload"])
        fields = flatten_dict(payload)
    except json.JSONDecodeError:
        fields = {"value": str(data["payload"].decode("utf-8"))}
    data = {
        "measurement": data["name"],
        "fields": fields
    }
    try:
        data["time"] = int(data["time"])
    except:
        pass

    return data


def simplify_serialized_data(data):
    try:
        simplified_data = {}

        for table in data:
            if table['name'] == "data":
                simplified_data["name"] = f"{table['measurement']}#{table['field']}"
                simplified_data["values"] = table["values"]
            else:
                simplified_data["sum"] = table["values"][0]["value"]

        return simplified_data
    except Exception as e:
        logger.warning(f"Failed to simplify the data, returning [{e}]")
        return data


def get_sum_from_serialized_data(data):
    logger.debug(data)
    try:
        for table in data:
            if table["name"] == "sum":
                return table["values"][0]["value"]
    except Exception as e:
        logger.warning(f"This data doesn't have sum, returning -1 [{e}]")
        return -1



# Companion Class
class MqttTopic:
    SENSORS = "homesolar/sensors/#"
    ACTUATORS = "homesolar/actuators/#"
    CLIENT = "homesolar/app/clients/#"
    REQUEST = "homesolar/app/clients/+/request"
    RESPONSE = "homesolar/app/clients/+/response"
    SUMMARY = "homesolar/summary"
    BATTERY = "homesolar/battery"




from datetime import datetime, timedelta, date
from time import time
from json import JSONEncoder


class DateTimeEncoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (datetime, date)):
            return obj.timestamp()


def stringify_timestamp(timestamp):
    return datetime.utcfromtimestamp(timestamp).strftime('%Y-%m-%dT%H:%M:%SZ')


def get_today():
    return datetime.fromtimestamp(time()).replace(hour=0, minute=0, second=0).timestamp()


def get_next_day(timestamp):
    return (datetime.utcfromtimestamp(timestamp) + timedelta(days=1)).strftime('%Y-%m-%dT%H:%M:%SZ')


def get_first_day_of_month(timestamp):
    return datetime.utcfromtimestamp(timestamp).replace(day=1).strftime('%Y-%m-%dT%H:%M:%SZ')


def get_last_day_of_month(timestamp):
    next_month = datetime.utcfromtimestamp(timestamp).replace(day=28) + timedelta(days=4)
    return (next_month - timedelta(days=next_month.day)).strftime('%Y-%m-%dT%H:%M:%SZ')


def get_first_day_of_year(timestamp):
    return datetime.utcfromtimestamp(timestamp).replace(day=1, month=1).strftime('%Y-%m-%dT%H:%M:%SZ')


def get_last_day_of_year(timestamp):
    return datetime.utcfromtimestamp(timestamp).replace(day=31, month=12).strftime('%Y-%m-%dT%H:%M:%SZ')


def get_date_pair(date, timescale):
    if timescale == "MONTH":
        start_time = get_first_day_of_month(date)
        stop_time = get_last_day_of_month(date)
    elif timescale == "YEAR":
        start_time = get_first_day_of_year(date)
        stop_time = get_last_day_of_year(date)
    else:
        start_time = stringify_timestamp(date)
        stop_time = get_next_day(date)

    return start_time, stop_time
